/*
	Copyright (C) 1995-2001 AROS - The Amiga Research OS
	$Id: intuition_init.c,v 1.23 2022/02/18 12:50:29 jacadcaps Exp $
 
	Desc: Library header for intuition
	Lang: English
*/

/****************************************************************************************/

#define AROS_ALMOST_COMPATIBLE

#define INIT AROS_SLIB_ENTRY(init,Intuition)

#include <string.h>
#include <exec/lists.h>
#include <exec/resident.h>
#include <exec/memory.h>
#include <exec/execbase.h>
#include <exec/alerts.h>
#include <proto/graphics.h>
#include <proto/exec.h>
#include <proto/intuition.h>
#include <proto/alib.h>
#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dostags.h>
#include <proto/dos.h>
#include <prefs/serial.h>
#include <prefs/icontrol.h>
//#include <proto/arossupport.h>
#include <devices/input.h>
#ifndef INTUITION_CLASSES_H
#   include <intuition/classes.h>
#endif
#include <intuition/pointerclass.h>
#ifndef UTILITY_HOOKS_H
#   include <utility/hooks.h>
#endif
#include <utility/utility.h>
#include <aros/asmcall.h>
#include "libdefs.h"
#ifdef __MORPHOS__
#ifdef VERSION_STRING
#undef VERSION_STRING
#endif
#ifdef VERSION_NUMBER
#undef VERSION_NUMBER
#endif
#ifdef REVISION_NUMBER
#undef REVISION_NUMBER
#endif
#include "intuition_rev.h"
#define VERSION_STRING VERSTAG
#define VERSION_NUMBER VERSION
#define REVISION_NUMBER REVISION
#endif
#include "intuition_intern.h"
#include "strgadgets.h" /* To get GlobalEditFunc prototype */
#include "inputhandler.h"
#include "menutask.h"
#include "menus.h"

#ifdef SKINS
#include "transplayers.h"
#include "smallmenu.h"
#include "pointers.h"
#include "specialclasses.h"
#include "blankertask.h"
#include "screenmanager.h"
#include "intuition_customizesupport.h"
#include "layermanager.h"
#endif

#define DEBUG 0
#include <aros/debug.h>

#ifdef INTUITION_NOTIFY_SUPPORT
/* screennotify/notifyintuition init routines from notify.c */
struct Library *sn_Init(struct IntuitionBase *IntuitionBase);
struct Library *ni_Init(struct IntuitionBase *IntuitionBase);
#endif

#undef DEBUG_INIT
#define DEBUG_INIT(x) ;
#define DEBUG_OPEN(x) ;

/****************************************************************************************/

static const char name[];
static const char version[];
static const APTR inittabl[4];
extern const void *const LIBFUNCTABLE[];

LIBBASETYPEPTR INIT();

extern const char LIBEND;

/* There has to be a better way... */

#if INTERNAL_BOOPSI

AROS_UFP3(ULONG, rootDispatcher,
		  AROS_UFPA(Class *,  cl,  A0),
		  AROS_UFPA(Object *, obj, A2),
		  AROS_UFPA(Msg,      msg, A1)
		 );

#endif

struct IClass *InitICClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitModelClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitImageClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitFrameIClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitSysIClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitFillRectClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitGadgetClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitButtonGClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitFrButtonClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitPropGClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitStrGClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitGroupGClass (LIBBASETYPEPTR LIBBASE);

struct IClass *InitMenuBarLabelClass (LIBBASETYPEPTR LIBBASE);

//struct IClass *InitDragBarClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitSizeButtonClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitPointerClass (LIBBASETYPEPTR LIBBASE);
struct IClass *InitWindowSysIButtonClass(LIBBASETYPEPTR LIBBASE);


int Intuition_entry(void)
{
	/* If the library was executed by accident return error code. */
	return -1;
}

const struct Resident Intuition_resident=
	{
	    RTC_MATCHWORD,
	    (struct Resident *)&Intuition_resident,
#ifdef __MORPHOS__
	    (APTR) (&Intuition_resident + 2),
	    RTF_AUTOINIT|RTF_COLDSTART|RTF_PPC|RTF_EXTENDED,
#else
	    (APTR)&LIBEND,
	    RTF_AUTOINIT|RTF_COLDSTART,
#endif
	    VERSION_NUMBER,
	    NT_LIBRARY,
	    10,
	    (char *)name,
	    (char *)VSTRING,
	    (ULONG *)inittabl
#ifdef __MORPHOS__
	    ,REVISION_NUMBER,
	    NULL   /* tags */
#endif
	};

static const char name[]=INTUITIONNAME;

static const char version[]=VERSION_STRING;

static const APTR inittabl[4]=
	{
		(APTR)sizeof(struct IntIntuitionBase),
		(APTR)LIBFUNCTABLE,
		NULL,
		&INIT
	};

#ifdef __MORPHOS__
const int __abox__ = 1;
#endif

/****************************************************************************************/

extern const ULONG coltab[];

#ifdef __MORPHOS__
struct LIBBASETYPE *LIB_init(struct LIBBASETYPE *LIBBASE, BPTR segList, struct ExecBase *sysBase)
{
#else
AROS_UFH3(LIBBASETYPEPTR, AROS_SLIB_ENTRY(init,Intuition),
		  AROS_UFHA(LIBBASETYPEPTR,	LIBBASE,    D0),
		  AROS_UFHA(BPTR,		segList,    A0),
		  AROS_UFHA(struct ExecBase *,	sysBase,    A6)
		 )
#endif
	{
		AROS_LIBFUNC_INIT
		SysBase = sysBase;

		DEBUG_INIT(dprintf("LIB_Init: base 0x%lx\n", (ULONG) LIBBASE));

#warning "FIXME: This libInit is all broken if something should fail, but do we care?"
#warning "FIXME: If something fails we're screwed anyway..."

#ifdef SKINS
		InitComplain(LIBBASE);
#endif

		/*  We have to open this here, but it doesn't do any allocations,
		so it shouldn't fail...
		*/

#if !INTERNAL_BOOPSI
		if (!(BOOPSIBase = OpenLibrary("boopsi.library", 0)))
		{
			/* Intuition couldn't open unknown library */
			Alert(AT_DeadEnd | AN_Intuition | AG_OpenLib | AO_Unknown);
			return NULL;
		}
#endif

		/* Create semaphore and initialize it */
		GetPrivIBase(LIBBASE)->IBaseLock = AllocMem (sizeof(struct SignalSemaphore), MEMF_PUBLIC|MEMF_CLEAR);

		if (!GetPrivIBase(LIBBASE)->IBaseLock)
			return NULL;

		InitSemaphore(GetPrivIBase(LIBBASE)->IBaseLock);
		InitSemaphore(&GetPrivIBase(LIBBASE)->LIBOpenLock);

		/* Initialize global stringgadget edit hook */
		GetPrivIBase(LIBBASE)->DefaultEditHook.h_Entry	= (APTR)AROS_ASMSYMNAME(GlobalEditFunc);
		GetPrivIBase(LIBBASE)->DefaultEditHook.h_SubEntry	= NULL;
		GetPrivIBase(LIBBASE)->DefaultEditHook.h_Data	= LIBBASE;

		GetPrivIBase(LIBBASE)->GlobalEditHook = &(GetPrivIBase(LIBBASE)->DefaultEditHook);

		GetPrivIBase(LIBBASE)->DefaultPubScreen = NULL;
		NEWLIST(&GetPrivIBase(LIBBASE)->PubScreenList);
		NEWLIST(&GetPrivIBase(LIBBASE)->DeadPorts);
		InitSemaphore(&GetPrivIBase(LIBBASE)->PubScrListLock);

		InitSemaphore(&GetPrivIBase(LIBBASE)->GadgetLock);
		InitSemaphore(&GetPrivIBase(LIBBASE)->MenuLock);
		InitSemaphore(&GetPrivIBase(LIBBASE)->WindowLock);
		InitSemaphore(&GetPrivIBase(LIBBASE)->IntuiActionLock);
		InitSemaphore(&GetPrivIBase(LIBBASE)->InputHandlerLock);

#ifdef SKINS
		InitSemaphore(&GetPrivIBase(LIBBASE)->DataTypesSem);
		GetPrivIBase(LIBBASE)->DataTypesBase = 0;

		GetPrivIBase(LIBBASE)->transphook.h_Data = (APTR)GetPrivIBase(LIBBASE);
		GetPrivIBase(LIBBASE)->transphook.h_Entry = (HOOKFUNC)HookEntry;
		GetPrivIBase(LIBBASE)->transphook.h_SubEntry = (HOOKFUNC)WindowTranspFunc;

		GetPrivIBase(LIBBASE)->notransphook.h_Data = (APTR)GetPrivIBase(LIBBASE);
		GetPrivIBase(LIBBASE)->notransphook.h_Entry = (HOOKFUNC)HookEntry;
		GetPrivIBase(LIBBASE)->notransphook.h_SubEntry = (HOOKFUNC)WindowNoTranspFunc;
#endif

#ifdef __MORPHOS__
		memset(GetPrivIBase(LIBBASE)->Pad, 0xee, sizeof(GetPrivIBase(LIBBASE)->Pad));
		GetPrivIBase(LIBBASE)->SystemRequestTitle = "System Request";
		GetPrivIBase(LIBBASE)->WorkbenchTitle = "Ambient Screen";

		/*
		 * Setup the default pens to the default
		 * colors so that screens have proper color
		 * even before IPrefs is loaded.
		 */
		{
			ULONG i;
			struct Color32 *p;

			p = GetPrivIBase(LIBBASE)->Colors;

			#ifdef SKINS
			for (i = 0; i < 8; i++)
			#else
			for (i = 0; i < 16; i++)
			#endif
			{
				p[i].red   = coltab[i * 3 + 1];
				p[i].green = coltab[i * 3 + 2];
				p[i].blue  = coltab[i * 3 + 3];
			}
		}

#endif

#if INTERNAL_BOOPSI
		InitSemaphore(&GetPrivIBase(LIBBASE)->ClassListLock);
		NEWLIST(&GetPrivIBase(LIBBASE)->ClassList);

		/* Setup root class */

		GetPrivIBase(LIBBASE)->RootClass.cl_Dispatcher.h_Entry = (APTR)AROS_ASMSYMNAME(rootDispatcher);
		GetPrivIBase(LIBBASE)->RootClass.cl_ID                 = (ClassID)ROOTCLASS;
		GetPrivIBase(LIBBASE)->RootClass.cl_UserData           = (IPTR)LIBBASE;

		DEBUG_INIT(dprintf("LIB_Init: create rootclass\n"));
		AddClass((struct IClass *)(&(GetPrivIBase(LIBBASE)->RootClass)));
#endif

		/* Add all other classes */

#if INTERNAL_BOOPSI
		DEBUG_INIT(dprintf("LIB_Init: create icclass\n"));
		InitICClass (LIBBASE);		/* After ROOTCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create modelclass\n"));
		InitModelClass (LIBBASE);		/* After ICCLASS	*/
#endif
		DEBUG_INIT(dprintf("LIB_Init: create imageclass\n"));
		InitImageClass (LIBBASE);		/* After ROOTCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create frameiclass\n"));
		InitFrameIClass (LIBBASE);		/* After IMAGECLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create sysiclass\n"));
		InitSysIClass (LIBBASE);		/* After IMAGECLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create fillrectclass\n"));
		InitFillRectClass (LIBBASE);	/* After IMAGECLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create itexticlass\n"));
		InitITextIClass (LIBBASE);		/* After IMAGECLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create gadgetclass\n"));
		InitGadgetClass (LIBBASE);		/* After ROOTCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create buttonclass\n"));
		InitButtonGClass (LIBBASE);		/* After GADGETCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create frbuttonclass\n"));
		InitFrButtonClass (LIBBASE);	/* After BUTTONGCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create propgclass\n"));
		GetPrivIBase(LIBBASE)->propgclass = InitPropGClass (LIBBASE); /* After GADGETCLASS	  */
		DEBUG_INIT(dprintf("LIB_Init: create strgclass\n"));
		InitStrGClass (LIBBASE);		/* After GADGETCLASS	*/
		DEBUG_INIT(dprintf("LIB_Init: create groupgclass\n"));
		InitGroupGClass (LIBBASE);		/* After GADGETCLASS	*/

#ifdef __MORPHOS__
		InitWindowClass(LIBBASE);
		InitScreenClass(LIBBASE);
		InitMonitorClass(LIBBASE);
		InitMonitorModeClass(LIBBASE);
		InitIntuiMessageClass(LIBBASE);
#endif

		#if 0
		GetPrivIBase(LIBBASE)->mosmenuclass = InitMuiMenuClass(LIBBASE);
		#endif

		DEBUG_INIT(dprintf("LIB_Init: create menubarlabelclass\n"));
		InitMenuBarLabelClass (LIBBASE); /* After IMAGECLASS */

		DEBUG_INIT(dprintf("LIB_Init: create dragbarclass\n"));
		//GetPrivIBase(LIBBASE)->dragbarclass = InitDragBarClass (LIBBASE); /* After GADGETCLASS */
		//if (!GetPrivIBase(LIBBASE)->dragbarclass)
		//	  return NULL;

		DEBUG_INIT(dprintf("LIB_Init: create sizebuttonclass\n"));
		GetPrivIBase(LIBBASE)->sizebuttonclass = InitSizeButtonClass (LIBBASE); /* After GADGETCLASS */
		if (!GetPrivIBase(LIBBASE)->sizebuttonclass)
			return NULL;

		GetPrivIBase(LIBBASE)->windowsysiclass = InitWindowSysIButtonClass(LIBBASE);
		if (!GetPrivIBase(LIBBASE)->windowsysiclass)
			return NULL;


		DEBUG_INIT(dprintf("LIB_Init: create pointerclass\n"));
		GetPrivIBase(LIBBASE)->pointerclass = InitPointerClass (LIBBASE);
		if (!GetPrivIBase(LIBBASE)->pointerclass)
			return NULL;

		DEBUG_INIT(dprintf("LIB_Init: load default preferences\n"));
		LoadDefaultPreferences(LIBBASE);

#ifdef SKINS
		if (!GetPrivIBase(LIBBASE)->SmallMenuPool)
		{
			if (!(GetPrivIBase(LIBBASE)->SmallMenuPool = CreatePool(MEMF_SEM_PROTECTED,(sizeof (struct SmallMenuEntry))*20,(sizeof (struct SmallMenuEntry))*20))) return NULL;
		}
#endif

		/* do NOT use semprot here, Alloc/FreeIntuiMessage is protected anyway */
		if (!(GetPrivIBase(LIBBASE)->IDCMPPool = CreatePool(/*MEMF_SEM_PROTECTED*/MEMF_PUBLIC,(sizeof (struct IntIntuiMessage)) * 100,sizeof (struct IntIntuiMessage)))) return NULL;
		InitSemaphore(&GetPrivIBase(LIBBASE)->IDCMPCacheLock);
		NEWLIST(&GetPrivIBase(LIBBASE)->IDCMPCacheList);

		if (!(GetPrivIBase(LIBBASE)->MenutaskPool = CreatePool(MEMF_SEM_PROTECTED|MEMF_CLEAR,(sizeof (struct MenuMessage)) * 100,sizeof (struct MenuMessage)))) return NULL;

		LIBBASE->LibNode.lib_Revision = REVISION_NUMBER;

#ifdef SKINS
		strcpy(IBase->IControlExtensions.ice_ClockFormat,"%X");
		IBase->IControlPrefs.ic_Flags = ICF_SCREENACTIVATION | ICF_DEFPUBSCREEN | ICF_STICKYMENUS |
			ICF_STICKYDELAY | ICF_OPAQUEMOVE | ICF_BORDERDRAGGING | ICF_OPAQUERESIZE;
		int_getdefaultiactions(IntuitionBase);
		
		//blanker default settings
		strcpy(IBase->BlankerPrefs.BlankerName,"butterflies.btd");
		IBase->BlankerPrefs.EventFlags = 0xFFFFFFFF & ~(0);
		IBase->BlankerPrefs.OffFlags = 0xFFFFFFFF & ~(0);
		IBase->BlankerPrefs.DPMSOnTime = 18000;
		IBase->BlankerPrefs.BlankTime = 6000;

		IBase->InputMousePrefs.ip_DoubleClick.tv_micro = 500000;
		IBase->InputMousePrefs.ip_PointerTicks = 2000;
		IBase->InputMousePrefs.ip_Flags = IPMF_MOUSESCALING;
		IBase->InputMousePrefs.ip_PointerBlank.tv_secs = 30;

		IBase->LayerSettings.AllowDoubleBuffered = TRUE;
		IBase->LayerSettings.AllowTripleBuffered = TRUE;
		IBase->LayerSettings.IgnoreMemoryChecks = FALSE;
		IBase->LayerSettings.LoadBalancing = TRUE;

		IBase->LayerManagerSettings.OpenWindowEffectSpeed = 0;
		IBase->LayerManagerSettings.CloseWindowEffectSpeed = 0;
		IBase->LayerManagerSettings.ActiveWindowEffectSpeed = 0;
		IBase->LayerManagerSettings.InactiveWindowOpacity = 100;
		IBase->LayerManagerSettings.InactiveWindowEffectSpeed = 0;

		strcpy(IBase->Codepage,"ISO-8859-1");
#endif

		IBase->SerialPrefs.sp_Unit0Map = 0;
		IBase->SerialPrefs.sp_BaudRate = 115200;
		IBase->SerialPrefs.sp_InputBuffer = 0xFFFF;
		IBase->SerialPrefs.sp_OutputBuffer = 0xFFFF;
		IBase->SerialPrefs.sp_InputHandshake = HSHAKE_XON;
		IBase->SerialPrefs.sp_OutputHandshake = HSHAKE_XON;
		IBase->SerialPrefs.sp_Parity = PARITY_NONE;
		IBase->SerialPrefs.sp_BitsPerChar = 8;
		IBase->SerialPrefs.sp_StopBits = 1;

		LIBBASE->LibNode.lib_Node.ln_Pri = 8;

		DEBUG_INIT(dprintf("LIB_Init: done\n"));
		/* You would return NULL if the init failed */
		return LIBBASE;
		AROS_LIBFUNC_EXIT
	}

	/****************************************************************************************/

static
LONG DoTaskIO(struct IntuitionBase *IntuitionBase, struct IORequest *ioreq)
{
	LONG ret;
	struct MsgPort *mp = &GetPrivIBase(LIBBASE)->InputMP;

	mp->mp_Flags  = PA_SIGNAL;
	mp->mp_SigBit = AllocSignal(-1);
	if ((BYTE) mp->mp_SigBit == (BYTE) -1)
	{
		mp->mp_SigBit = SIGB_INTUITION;
		SetSignal(0, SIGF_INTUITION);
	}
	mp->mp_SigTask = FindTask(NULL);

	/* mn_ReplyPort is already set by CreateIORequest! */
	/* ioreq->io_Message.mn_ReplyPort = mp; */
	ret = DoIO(ioreq);

	mp->mp_Flags = PA_IGNORE;

	if (mp->mp_SigBit != SIGB_INTUITION)
	{
		FreeSignal(mp->mp_SigBit);
	}

	return ret;
}


	AROS_LH1(LIBBASETYPEPTR, open,
	         AROS_LHA(ULONG, version, D0),
	         LIBBASETYPEPTR, LIBBASE, 1, Intuition)
	{
		AROS_LIBFUNC_INIT

		DEBUG_OPEN(dprintf("LIB_Open: base 0x%lx\n", LIBBASE));

		/* Keep compiler happy */
		version=version;

		DEBUG_OPEN(dprintf("LIB_Open: ObtainSemaphore LIBOpenLock\n"));

		ObtainSemaphore(&GetPrivIBase(LIBBASE)->LIBOpenLock);

		DEBUG_OPEN(dprintf("LIB_Open: ObtainSemaphore LIBOpenLock done\n"));

		/* Open the input device */

		if (!GetPrivIBase(LIBBASE)->InputIO)
		{
			struct MsgPort *mp = &GetPrivIBase(LIBBASE)->InputMP;

			mp->mp_Node.ln_Type = NT_MSGPORT;
			mp->mp_Flags        = PA_IGNORE;
			mp->mp_SigBit       = -1;
			mp->mp_SigTask      = NULL;
			NEWLIST(&mp->mp_MsgList);

			if (!(GetPrivIBase(LIBBASE)->InputIO = (struct IOStdReq *)
			                                       CreateIORequest(mp, sizeof (struct IOStdReq))) )
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't create iorequest\n"));
				goto fail;
			}
		}

		if (!GetPrivIBase(LIBBASE)->InputDeviceOpen)
		{
			if (!OpenDevice("input.device", -1, (struct IORequest *)GetPrivIBase(LIBBASE)->InputIO, 0))
			{
				GetPrivIBase(LIBBASE)->InputDeviceOpen = TRUE;
				InputBase = (struct Library *)GetPrivIBase(LIBBASE)->InputIO->io_Device;
			}
			else
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open input.device\n"));
				goto fail;
			}
		}

		if (!UtilityBase)
		{
#ifdef SKINS
			if (!(UtilityBase = (void *)ComplainOpenLibrary(COMPLAIN_UTILITY,TRUE,IntuitionBase)))
#else
			if (!(UtilityBase = (void *)OpenLibrary (UTILITYNAME, 39)) )
#endif
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open utility.library\n"));
				goto fail;
			}
		}

		if (!GfxBase)
		{
			struct ViewExtra *ve;

#ifdef SKINS
			if (!(GfxBase = (void *)ComplainOpenLibrary(COMPLAIN_GRAPHICS,TRUE,IntuitionBase)))
#else
			if (!(GfxBase = (void *)OpenLibrary (GRAPHICSNAME, 39)) )
#endif
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open graphics.library\n"));
				goto fail;
			}

#ifdef __MORPHOS__
			if (!(ve = GfxNew(VIEW_EXTRA_TYPE)))
			{
				GfxBase = NULL;
				DEBUG_OPEN(dprintf("LIB_Open: can't create view extra\n"));
				goto fail;
			}

			InitView(&IntuitionBase->ViewLord);

			GfxAssociate(&IntuitionBase->ViewLord, ve);

			GetPrivIBase(LIBBASE)->ViewLordExtra = ve;

			GetPrivIBase(LIBBASE)->SpriteNum = -1;
#endif
			GetPrivIBase(LIBBASE)->ScreenFont = GfxBase->DefaultFont;

			{
				struct TextAttr textattr = {"topaz.font",8,0,FPF_ROMFONT};
				if (!(GetPrivIBase(LIBBASE)->TopazFont))
				GetPrivIBase(LIBBASE)->TopazFont = OpenFont(&textattr);
			}
		}

		if (!LayersBase)
		{
#ifdef SKINS
			if (!(LayersBase = (void *)ComplainOpenLibrary(COMPLAIN_LAYERS,TRUE,IntuitionBase)))
#else
			if (!(LayersBase = (void *)OpenLibrary ("layers.library", 39)) )
#endif
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open layers.library\n"));
				goto fail;
			}
		}

		/* cgx moved to OpenScreen!*/

		if (!KeymapBase)
		{
#ifdef SKINS
			if (!(KeymapBase = ComplainOpenLibrary(COMPLAIN_KEYMAP,TRUE,IntuitionBase)))
#else
			if (!(KeymapBase = OpenLibrary ("keymap.library", 39)) )
#endif
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open keymap.library\n"));
				goto fail;
			}
			DEBUG_OPEN(dprintf("LIB_Open: opened KeymapBase 0x%lx\n", KeymapBase));
		}

		if (!TimerBase)
		{
			struct MsgPort         TimerMP;
			struct timerequest     *TimerIO;

			TimerMP.mp_Node.ln_Type = NT_MSGPORT;
			TimerMP.mp_Flags        = PA_IGNORE;
			TimerMP.mp_SigBit       = -1;
			TimerMP.mp_SigTask      = NULL;
			NEWLIST(&TimerMP.mp_MsgList);

			if (!(TimerIO = (struct timerequest *)CreateIORequest(&TimerMP, sizeof(struct timerequest))))
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't create timer ioreq\n"));
				goto fail;
			}

			if (OpenDevice(TIMERNAME,UNIT_VBLANK, (struct IORequest *)TimerIO,0))
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't open timer.device\n"));
				DeleteIORequest((struct IORequest *)TimerIO);
				goto fail;
			}

			TimerBase = (struct Library *)TimerIO->tr_node.io_Device;

			CloseDevice((struct IORequest *)TimerIO);
			DeleteIORequest((struct IORequest *)TimerIO);

			SetPrefs(&IBase->DefaultPreferences, sizeof(struct Preferences), FALSE);

#ifdef SKINS
			if (!(CreateDefaultPointers(IntuitionBase))) goto fail;
#else
#warning: fixme: no pointers for non-skin build
#endif
		}

		if (!DOSBase)
		{
#ifdef SKINS
			DOSBase = ComplainOpenLibrary(COMPLAIN_DOS,TRUE,IntuitionBase);
#else
			DOSBase = OpenLibrary("dos.library", 50);
#endif

#ifdef SKINS
			if (DOSBase)
			{
				DEBUG_OPEN(dprintf("LIB_Open: create menu handler task\n"));
				/* FIXME: no cleanup routines for MenuHandler task */
				if (!InitDefaultMenuHandler(IntuitionBase))
				{
					DEBUG_OPEN(dprintf("LIB_Open: create menu handler task failed\n"));
					goto fail;
				}
				DEBUG_OPEN(dprintf("LIB_Open: InitSkinManager\n"));
				InitSkinManager(IntuitionBase);
				DEBUG_OPEN(dprintf("LIB_Open: InitSkinManager done\n"));
				DEBUG_OPEN(dprintf("LIB_Open: InitScreenManager\n"));
				InitScreenManager(IntuitionBase);
				DEBUG_OPEN(dprintf("LIB_Open: InitScreenManager done\n"));
				InitLayerManager(LIBBASE);
			} else {
				goto fail;
			}
		}
#endif
		if (!GetPrivIBase(LIBBASE)->InputHandler)
		{
			D(bug("Initializing inputhandler\n"));
			if ( !(GetPrivIBase(LIBBASE)->InputHandler = InitIIH(LIBBASE)) )
			{
				DEBUG_OPEN(dprintf("LIB_Open: can't init input handler\n"));
				goto fail;
			}

			D(bug("Adding inputhandler\n"));
			GetPrivIBase(LIBBASE)->InputIO->io_Data = (APTR)GetPrivIBase(LIBBASE)->InputHandler;
			GetPrivIBase(LIBBASE)->InputIO->io_Command = IND_ADDHANDLER;

			D(bug("Calling DoTaskIO\n"));
			DoTaskIO(IntuitionBase, (struct IORequest *)GetPrivIBase(LIBBASE)->InputIO);
			D(bug("Calling DoTaskIO done\n"));
		}

#ifdef __MORPHOS__
		if (!GetPrivIBase(LIBBASE)->RamlibTask)
		{
			GetPrivIBase(LIBBASE)->RamlibTask = FindTask("ramlib");
			DEBUG_OPEN(dprintf("LIB_Open: RamlibTask 0x%lx\n",GetPrivIBase(LIBBASE)->RamlibTask));
		}
#endif


#ifdef INTUITION_NOTIFY_SUPPORT
		/* Add screennotify.library base if not there yet - Piru
		 */
		if (!GetPrivIBase(LIBBASE)->ScreenNotifyBase)
		{
			DEBUG_OPEN(dprintf("LIB_Open: open ScreenNotify\n"));
			GetPrivIBase(LIBBASE)->ScreenNotifyBase = sn_Init(IntuitionBase);
			DEBUG_OPEN(dprintf("LIB_Open: ScreenNotifyBase 0x%lx\n",GetPrivIBase(LIBBASE)->ScreenNotifyBase));
		}

#if 0 /* not finished yet - Piru */
		/* Add notifyintuition.library base if not there yet - Piru
		 */
		if (!GetPrivIBase(LIBBASE)->NotifyIntuitionBase)
		{
			DEBUG_OPEN(dprintf("LIB_Open: open NotifyIntuitionBase\n"));
			GetPrivIBase(LIBBASE)->NotifyIntuitionBase = ni_Init(IntuitionBase);
			DEBUG_OPEN(dprintf("LIB_Open: NotifyIntuitionBase 0x%lx\n",GetPrivIBase(LIBBASE)->NotifyIntuitionBase));
		}
#endif
#endif

		DEBUG_OPEN(dprintf("LIB_Open: increase OpenCnt\n"));
		/* I have one more opener. */
		Forbid();
		LIBBASE->LibNode.lib_OpenCnt++;
		LIBBASE->LibNode.lib_Flags&=~LIBF_DELEXP;
		Permit();

		DEBUG_OPEN(dprintf("LIB_Open: release LIBOpenLock\n"));
		ReleaseSemaphore(&GetPrivIBase(LIBBASE)->LIBOpenLock);
		DEBUG_OPEN(dprintf("LIB_Open: done, all ok\n"));

		/* You would return NULL if the open failed. */
		return LIBBASE;

fail:
		ReleaseSemaphore(&GetPrivIBase(LIBBASE)->LIBOpenLock);
		DEBUG_OPEN(dprintf("LIB_Open: failed\n"));
		return NULL;
		
		AROS_LIBFUNC_EXIT
	}

	/****************************************************************************************/

	AROS_LH0(BPTR, close,
	         LIBBASETYPEPTR, LIBBASE, 2, Intuition)
	{
		AROS_LIBFUNC_INIT
		BPTR ret = 0;

		/* I have one fewer opener. */
		Forbid();
		if (!--LIBBASE->LibNode.lib_OpenCnt)
		{

#if 0 /* intuition_driver stuff is dead */
			intui_close(LIBBASE);
#endif

#if 0
			/* Delayed expunge pending? */
			if (LIBBASE->LibNode.lib_Flags & LIBF_DELEXP)
				/* Then expunge the library */
				ret = expunge();
#endif
		}
		Permit();
		return ret;
		AROS_LIBFUNC_EXIT
	}

	/****************************************************************************************/

	AROS_LH0(BPTR, expunge,
	         LIBBASETYPEPTR, LIBBASE, 3, Intuition)
	{
		AROS_LIBFUNC_INIT
#if 1
		LIBBASE = LIBBASE;      /* shut up the compiler */
#else
		/* Test for openers. */
		Forbid();
		if (LIBBASE->LibNode.lib_OpenCnt)
		{
			/* Set the delayed expunge flag and return. */
			LIBBASE->LibNode.lib_Flags |= LIBF_DELEXP;
			Permit();
			return 0;
		}
		Permit();

		/* We are no longer patching DOS */
		((struct DosLibrary *)DOSBase)->dl_IntuitionBase = NULL;

		/* Reset the old DisplayError() function */
		SetFunction(DOSBase, -81*LIB_VECTSIZE,
		            GetPrivIBase(LIBBASE)->OldDisplayErrorFunc);

		/* Free unecessary memory. */

		/* The WB screen is opened in ./lateintuiinit.c */
		if (GetPrivIBase(LIBBASE)->WorkBench)
			CloseScreen (GetPrivIBase(LIBBASE)->WorkBench);

		if (KeymapBase)
			CloseLibrary (KeymapBase);

		if (UtilityBase)
			CloseLibrary ((struct Library *)UtilityBase);

		if (GfxBase)
		{
#ifdef __MORPHOS__
			GfxFree(GetPrivIBase(LIBBASE)->ViewLordExtra);
#endif
			CloseLibrary ((struct Library *)GfxBase);
		}

		if (DOSBase)
			CloseLibrary(DOSBase);


		if (GetPrivIBase(LIBBASE)->InputHandler)
		{
			CleanupIIH(GetPrivIBase(LIBBASE)->InputHandler, LIBBASE);

			/* Remove inputandler */
			GetPrivIBase(LIBBASE)->InputIO->io_Data = (APTR)GetPrivIBase(LIBBASE)->InputHandler;
			GetPrivIBase(LIBBASE)->InputIO->io_Command = IND_REMHANDLER;
			DoTaskIO(IntuitionBase, (struct IORequest *)GetPrivIBase(LIBBASE)->InputIO);
		}


		if (GetPrivIBase(LIBBASE)->InputDeviceOpen)
			CloseDevice((struct IORequest *)GetPrivIBase(LIBBASE)->InputIO);

		if (GetPrivIBase(LIBBASE)->InputIO)
			DeleteIORequest((struct IORequest *)GetPrivIBase(LIBBASE)->InputIO);


#if 0 /* intuition_driver stuff is dead */
		/* Let the driver do the same */
		intui_expunge (LIBBASE);
#endif

#ifdef DISK_BASED /* Don't remove a ROM library */
		FreeImageClass ();

		/* Get rid of the library. Remove it from the list. */
		Remove(&LIBBASE->LibNode.lib_Node);

		/* Free the memory. */
		FreeMem((char *)LIBBASE-LIBBASE->LibNode.lib_NegSize,
		        LIBBASE->LibNode.lib_NegSize+LIBBASE->LibNode.lib_PosSize);
#endif
#endif
		return 0L;
		AROS_LIBFUNC_EXIT
	}

	/****************************************************************************************/

	AROS_LH0I(int, null,
	          LIBBASETYPEPTR, LIBBASE, 4, Intuition)
	{
		AROS_LIBFUNC_INIT
		return 0;
		AROS_LIBFUNC_EXIT
	}

	/****************************************************************************************/
